from TDhelper.generic.transformationType import transformation
from TDhelper.document.excel.FieldType import *
from types import FunctionType, MethodType, ModuleType
from openpyxl import load_workbook
import csv
import copy


class _AttributeOverride:
    def __init__(self, name, m_type):
        self._name = name
        self._type = m_type

    def __get__(self, instance, owen):
        return instance.__dict__[self._name]

    def __set__(self, instance, value):
        instance.__dict__[self._name] = transformation(value, self._type)

    def __delete__(self, instance):
        instance.__dict__.pop(self._name)


class Meta:
    file = None
    sheet = 'sheet1'
    extension = 'xlsx'


class modelMeta(type):
    def __new__(cls, name, bases, dct):
        attrs = {'mapping': {}, 'Meta': Meta, '__exit__': __exit__, 'readLine': readLine,
                 'close': close, '__initExcelHandle__': __initExcelHandle__, 'rows': rows, 'getCount': getCount, '_translateMode': _translateMode, '__rowsCount__': 0}
        for name, value in dct.items():
            if (not isinstance(dct[name], type)) and (not isinstance(dct[name], FunctionType)):
                if not name.startswith('__'):
                    if isinstance(dct[name], FieldType):
                        attrs['mapping'][name] = value.bindCol
                        attrs[name] = _AttributeOverride(name, value.fieldType)
                    else:
                        raise Exception('field type must is FieldType.')
                else:
                    attrs[name] = value
            else:
                if isinstance(dct[name], type):
                    if name == 'Meta':
                        for attr_name in dct[name].__dict__:
                            if not attr_name.startswith('__'):
                                setattr(attrs['Meta'], attr_name,
                                        dct[name].__dict__[attr_name])
                else:
                    attrs[name] = value
        return super(modelMeta, cls).__new__(cls, name, bases, attrs)


def __initExcelHandle__(self):
    try:
        if self.Meta.file:
            m_extension = self.Meta.file.rsplit('.')[1]
            if m_extension == 'csv':
                self.Meta.extension = 'csv'
                self.__excelHandle__ = open(self.Meta.file)
                self.__sheetHandle__ = csv.reader(self.__excelHandle__)
                self.__excelHandle__ = None
                m_offset = 1
                for v in self.__sheetHandle__:
                    if v:
                        if m_offset < self.__record_offset__:
                            m_offset += 1
                            continue
                        self.__rows__.append(self._translateMode(v))
                    else:
                        break
                self.__rowsCount__ = len(self.__rows__)
            elif m_extension == 'xlsx' or m_extension == 'xls':
                self.Meta.extension = 'xlsx'
                self.__excelHandle__ = load_workbook(self.Meta.file)
                self.__sheetHandle__ = self.__excelHandle__[self.Meta.sheet]
                column = self.__sheetHandle__.max_column+1
                m_offset= 1
                tmp_handle= self.__sheetHandle__
                for m_rows in tmp_handle:
                    if m_offset < self.__record_offset__:
                        m_offset += 1
                        continue
                    self.__rows__.append(self._translateMode(m_rows))
                self.__rowsCount__ = len(self.__rows__)
                tmp_handle= None
            else:
                raise Exception('file extension is error.')
        else:
            raise Exception('meta file is None.')
    except Exception as e:
        raise e


def _translateMode(self, data=[]):
    m_extension = self.Meta.file.rsplit('.')[1].lower()
    if data:
        try:
            # 关闭IO，然后返回深拷贝对象解决数组里对象值引用引起的所有对象同样的值.
            # (没搞明白，为什么这里设为NONE了。外层的for循环还能继续下去)
            self.__sheetHandle__ = None
            self.__excelHandle__ = None
            for (name, value) in self.mapping.items():
                if value <= len(data):
                    if m_extension == 'csv':
                        attr_value = data[value-1]
                    elif m_extension == 'xlsx' or m_extension == 'xls':
                        attr_value= data[value-1].value
                    else:
                        raise Exception('file extension is error.')
                    if not attr_value:
                        attr_value = '0.00'
                    if attr_value == 'None':
                        attr_value = '0.00'
                    setattr(self, name, attr_value)
                else:
                    raise Exception('mapping error:(%s,%s)' % (name, value))
            return copy.deepcopy(self)
        except Exception as e:
            raise e
    else:
        raise Exception("translate data is none.")


def __exit__(self, exc_type, exc_value, exc_t):
    self.close()


def close(self):
    self.__excelHandle__ = None
    self.__sheetHandle__ = None


def getCount(self):
    return self.__rowsCount__


def rows(self):
    return self.__rows__


def readLine(self, lineOffset=0):
    if lineOffset >= self.getCount():
        return None
    self.__read_offset__= lineOffset
    return self.__rows__[lineOffset]



from multiprocessing import Process, Lock
from threading import Thread
from array import array
import queue
import copy
import time

def PoolsFunc(func):
    def wapper(*args:tuple, **kwargs):
        try:
            thread_pools= None
            thread_index= None
            if len(args) >= 2:
                thread_pools = args[0]
                thread_index = args[1]
            else:
                raise Exception('params error.')
            kwargs['handle']= thread_pools
            kwargs['thread-id']= thread_index
            if not isinstance(args[2],list):
                raise Exception("params args must is <list>. you given an <%s>." % type(args[2]).__name__)
            m_args=args[2]
            while thread_pools._processFlag:
                if not m_args:
                    break
                ret= func(*m_args, **kwargs)
                m_args= thread_pools.popArgs()
            thread_pools.threadComplete(thread_index)
            return ret
        except Exception as e:
            raise Exception(e)
    return wapper

class threadPools:
    def __init__(self, threadfunc, daemon=False, isJoin=False, maxSize=10):
        self.Daemon = daemon
        self.isJoin = isJoin 
        self._processFlag = True # 方法状态
        self.processLock= Lock() # 方法事务锁
        self._threadLockArgs = Lock() # 任务队列锁
        self._threadLockThreadFunc = Lock() # 线程队列锁
        self._activeThread = {} # 活跃线程
        self._list = queue.Queue(maxSize) # 线程队列
        self._args = [] # 任务队列
        for offset in range(0, maxSize):
            self._list.put((offset, threadfunc))

    def stop(self, waitProcessComplate= True):
        '''
            停止线程
            - params:
            -   waiProcessComplate: <boolen>, 是否等待线程任务结束, 默认值：True
        '''
        if not waitProcessComplate:
            self._processFlag = False
        else:
            while True:
                if self.getArgsLength()==0:
                    self._processFlag= False
                    break
                time.sleep(0.5)

    def restart(self):
        '''
            重启线程,未关闭的情况下
        '''
        self._processFlag = True

    def pushArgs(self, param: any):
        '''
            push线程任务队列.
            - params:
            -   param: <any>, 任务参数.
        '''
        self._threadLockArgs.acquire()
        self._args.append(param)
        self._threadLockArgs.release()
        self.popThread()

    def popArgs(self):
        '''
            弹出任务
        '''
        value = None
        self._threadLockArgs.acquire()
        if len(self._args) > 0:
            value = self._args[0]
            del self._args[0]
        self._threadLockArgs.release()
        return value

    def getArgsLength(self):
        '''
            获取任务队列长度
        '''
        value = 0
        self._threadLockArgs.acquire()
        value = len(self._args)
        self._threadLockArgs.release()
        return value

    def popThread(self):
        '''
            从线程队列里，弹出一个线程方法.
        '''
        try:
            if not self._list.empty():
                self._threadLockThreadFunc.acquire()
                m_thread = self._list.get()
                self._threadLockThreadFunc.release()
                if m_thread:
                    self._activeThread[m_thread[0]] = m_thread[1]
                    m_args = self.popArgs()
                    if m_args:
                        m_thread = Thread(target=self._activeThread[m_thread[0]], name='Thread%s' % str(
                            m_thread[0]), args=(self, m_thread[0], m_args,))
                        m_thread.setDaemon = self.Daemon
                        m_thread.start()
                        if self.isJoin:
                            m_thread.join()
                    else:
                        self.threadComplete(m_thread[0])
                else:
                    raise Exception('thread func is none.')
        except Exception as e:
            raise e

    def pushThread(self, m_thread: tuple):
        '''
            将一个方法装入线程队列.
        '''
        if not self._list.full():
            self._threadLockThreadFunc.acquire()
            self._list.put(m_thread)
            self._threadLockThreadFunc.release()
            if self.getArgsLength() > 0:
                self.popThread()
        else:
            raise Exception('thread pools is full.')

    def threadComplete(self, offset):
        '''
            线程任务执行完成,将方法重新放入队列
        '''
        m_thread = (offset, self._activeThread.pop(offset))
        self.pushThread(m_thread)






#!/usr/bin/env python3.6
# -*- encoding: utf-8 -*-
'''
@File    :   hip.py
@Time    :   2020/08/26 20:18:34
@Author  :   Tang Jing 
@Version :   1.0.0
@Contact :   yeihizhi@163.com
@License :   (C)Copyright 2020
@Desc    :   踝关节
'''

# here put the import lib
from TDhelper.robot.struct.base import joint
# code start


class ANKLE:
    def __init__(self, joints:[]= None):
        '''
            脚踝初始化
            - params:
            -   joints: <arrary>, 关节.
        '''
        self.joints ={}
        count= 0
        for v in joints:
            self.joints[count]= v
            count += 1



import queue
import copy
from multiprocessing import Process,Lock
from TDhelper.Event.Event import Event
class pools(Event):
    def __init__(self, maxSize=100):
        '''
            初始化

            - 参数:
                obj: 装入队列的对象, 一个定义好的对象.初始化程序会自动拷贝对象填满队列.
                maxSize: 队列长度
        '''
        super(pools, self).__init__()
        # 可用资源
        self.available_resources= 0
        self.size= maxSize
        self._list= queue.Queue(maxSize)
        self._lock=Lock()

    def push(self,value):
        try:
            self._lock.acquire()
            if not self._list.full():
                self.available_resources+= 1
                self._list.put(value)
        finally:
            self._lock.release()

    def pop(self):
        try:
            self._lock.acquire()
            if not self._list.empty():
                self.available_resources-= 1
                return self._list.get()
        finally:
            self._lock.release()

    def getQSize(self):
        return self._list.qsize()
        


import json
from TDhelper.network.http.REST_HTTP import GET, POST


class RPC:
    '''
        RPC. 此类只能配合webservice/rpc使用, 独立使用将会报错.
    '''

    def __init__(self, service_center_uri, secret):
        '''
            初始化
            - params:
            -   service_center_uri:<string>, 服务中心获取API接口URI.
            -   secret: <string>, 访问密钥
        '''
        self._access_token = secret
        self._sc_uri = service_center_uri
        self._apisTable = {}

    def _getApi(self, service, method):
        if service not in self._apisTable:
            # 远程获取
            m_heads = {"api-token": self._access_token}
            status, body = GET(uri=self._sc_uri+"rpc/?key="+service +
                               "&method="+method, time_out=15, http_headers=m_heads)
            if status == 200:
                self._apisTable[service] = json.loads(
                    str(body, encoding='utf-8'))
        # 从本地获取API配置
        if service in self._apisTable:
            if self._apisTable[service]['state'] == 200:
                return True, self._apisTable[service]['msg']
            else:
                return False, self._apisTable[service]['msg']
        else:
            return False, "can't found %s api." % service

    def register(self, service: dict, hosts: [], methods: []):
        '''
            注册服务
            - params:
            -   service: <dict>, 服务信息. formatter:{"name":"","description":"","key":"","httpProtocol":""}
            -   hosts: <[]>, 服务器信息. formatter:[{"host":"ip地址","port":端口}]
            -   methods: <[]>, 方法. formatter: [{"key":"方法索引","uri":"api url","method":"GET|POST|PUT|DELETE","version":"版本号","description":"描述", "params":[{"key":"参数名称","description":"描述","defaultValue":"默认值(调用不传参时默认值)"}]}]
        '''
        m_service_post_data = ''
        m_count = 0
        # 生成注册服务参数.
        for k, v in service.items():
            if k.lower() != 'description':
                if not v:
                    raise Exception("service '%s' value can't is none." % k)
            m_service_post_data += k+"="+str(v)
            m_count += 1
            if m_count < len(service):
                m_service_post_data += "&"
        # 注册服务基本信息.
        state, ret = POST(self._sc_uri+"services/",
                          post_data=bytes(m_service_post_data, encoding='utf-8'), time_out=15)
        if state == 200:
            if ret:
                ret = json.loads(str(ret, encoding='utf-8'))
                if ret['state'] == 200:
                    m_service_id = ret['msg']['id']
                    # 注册服务器信息.
                    for i in range(0, len(hosts)):
                        hosts[i]['service'] = m_service_id
                        #hosts[i]['state']= True
                        m_count = 0
                        m_service_hosts_post_data = ''
                        for k, v in hosts[i].items():
                            if v:
                                m_service_hosts_post_data += k+"="+str(v)
                                m_count += 1
                                if m_count < len(hosts[i]):
                                    m_service_hosts_post_data += "&"
                            else:
                                raise Exception(
                                    "register hosts {} can't is none." % k)
                        state, ret = POST(self._sc_uri+"hosts/", post_data=bytes(
                            m_service_hosts_post_data, encoding='utf-8'), time_out=15)
                        if state != 200:
                            return False, 'register service hosts failed. msg:{}' % str(ret, encoding='utf-8')
                        else:
                            ret= json.loads(str(ret, encoding='utf-8'))
                            if ret['state'] != 200:
                                m_msg = 'register service hosts failed. http code({}), msg:{}' % (ret[
                                    'state'], ret['msg'])
                                return False, m_msg
                    # 注册方法
                    for i in range(0, len(methods)):
                        # todo register method.
                        methods[i]['service'] = m_service_id
                        m_service_uri_post_data = ''
                        m_count = 0
                        for k, v in methods[i].items():
                            if k.lower() != 'params':
                                if k.lower() == 'key' or k.lower() == 'uri' or k.lower() == 'method':
                                    if not v:
                                        raise Exception(
                                            'methods (%s) is none.' % k.lower())
                                if k.lower()!='method':
                                    m_service_uri_post_data += k + "=" + str(v)
                                else:
                                    m_value= 0
                                    if v.upper()== u"GET":
                                        m_value= 1
                                    elif v.upper()== u"POST":
                                        m_value= 2
                                    elif v.upper()== u"PUT":
                                        m_value= 3
                                    elif v.upper()== u"DELETE":
                                        m_value= 4
                                    m_service_uri_post_data += k+ "=" + str(m_value)
                                m_count += 1
                                if 'params' in methods[i]:
                                    if m_count < len(methods[i])-1:
                                        m_service_uri_post_data += "&"
                                else:
                                    if m_count < len(methods[i]):
                                        m_service_uri_post_data += "&"
                        state, ret = POST(self._sc_uri+"uri/", post_data=bytes(
                            m_service_uri_post_data, encoding='utf8'), time_out=15)
                        if state != 200:
                            return False, 'register service methods failed. msg:{}' % str(ret, encoding='utf-8')
                        else:
                            ret= json.loads(str(ret, encoding= 'utf-8'))
                            if ret['state'] != 200:
                                m_msg = 'register service methods failed. http code(%d), msg:%s' % (ret['state'], ret['msg'])
                                return False, m_msg
                            else:
                                if ret:
                                    params = methods[i]['params']
                                    m_method_id = ret['msg']["id"]
                                    for param_offset in range(0,len(params)):
                                        params[param_offset]['serviceUri'] = m_method_id
                                        m_service_method_params_post_data = ''
                                        m_count = 0
                                        for k, v in params[param_offset].items():
                                            if not v:
                                                raise Exception(
                                                    'methods params (%s) is none.' % k.lower())
                                            m_service_method_params_post_data += k + \
                                                "=" + str(v)
                                            m_count += 1
                                            if m_count < len(params[i]):
                                                m_service_method_params_post_data += "&"
                                        state, ret = POST(
                                            self._sc_uri+"params/", post_data= bytes(m_service_method_params_post_data, encoding= 'utf-8'), time_out= 15)
                                else:
                                    if ret['state'] != 200:
                                        m_msg = 'register service hosts failed. http code(%d), msg:%s' % (ret[
                                            'state'], ret['msg'])
                                        return False, m_msg

                    return True, 'register service success.'
                else:
                    msg = "register service http error. http code(%s), msg:%s" % (ret[
                        'state'], ret['msg'])
                    return False, msg
            else:
                raise Exception(
                    "register return body is none, please checked.")
        else:
            msg = "register service http error. http code(%d), msg:%s" % (state, str(ret, encoding='utf-8'))
            return False, msg

    def call(self, service, method, **kwargs):
        state, ret = self._getApi(service, method)
        if state:
            try:
                m_uri = ret['uri']
                m_params = ret['params']
                m_method = ret['method']
                params_data = ''
                m_count = 0
                for param in m_params:
                    if param['key'] in kwargs:
                        params_data += param['key'] + \
                            '=' + str(kwargs[param['key']])
                    else:
                        params_data += param['key']+'=' + param['defaultValue']
                    m_count += 1
                    if m_count < len(kwargs):
                        params_data += "&"
                try:
                    if m_method == u"GET":
                        m_uri += "?" + params_data
                        state, ret = GET(m_uri, time_out=15)
                        if state == 200:
                            return {"httpCode": state, "data": json.loads(str(ret, encoding='utf-8'))}
                        else:
                            return {"httpCode": state, "data": str(ret, encoding='utf-8')}
                    elif m_method == u"POST":
                        state, ret = POST(m_uri, bytes(
                            params_data, encoding='utf-8'), time_out=15)
                        if state == 200:
                            return {"httpCode": state, "data": json.loads(str(ret, encoding='utf-8'))}
                        else:
                            return {"httpCode": state, "data": str(ret, encoding='utf-8')}
                    elif m_method == u"PUT":
                        # 还没有写PUT方法
                        raise Exception('urllib PUT方法还没写.')
                    elif m_method == u"DELETE":
                        # 还没有写DELETE方法
                        raise Exception("urllib delete 方法还没写.")
                    else:
                        raise Exception('restful method is error.')
                except Exception as e:
                    return {"httpCode": -1, "data": e}
            except Exception as e:
                raise e
        else:
            return {"httpCode": -1, "data": str(ret, encoding='utf-8')}


if __name__ == '__main__':
    m_rpc = RPC("http://127.0.0.1:8000/api/", "349304398403804983048034")
    print(m_rpc.register(
        service={"name": "test", "description": "测试添加服务",
                 "key": "gg", "httpProtocol": "http://"},
        hosts=[{"host": "127.0.0.1", "port": 8002}, {
            "host": "192.168.0.100", "port": 8000}],
        methods=[{"key": "test_method", "uri": "api/uri/", "method": "POST", "version": "1.0.0", "description": "测试添加方法", "params": [{"key": "test_method_p1", "description": "P1参数", "defaultValue": "0"}]}]))
    #print(m_rpc.call("remote_config","remote_config_register", **{"name":"测试远程配置服务注册","serivce_key":"test_rpc_register"}))



import urllib
import socket
from urllib import request
from urllib.error import HTTPError, URLError

def GET(uri:str, http_headers:dict= None, time_out:int= 1):
    """
    http GET method
    
    - Parameters:
        uri: an uri, it can be an domain or ip, type must is cls<str>.
        http_headers: set request's headers, default is None.
        time_out: time out for access remote uri , default value is 1 seconds.

    - Returs:
        stauts, body

        example: 
            
            200, <html><body>this is an example</body></html>
    """
    try:
        req= None
        if http_headers:
            req= request.Request(uri, headers= http_headers, method= 'GET')
        else:
            req= request.Request(uri, method= 'GET')
        with request.urlopen(req,timeout= time_out) as response:
            return response.getcode(),response.read()
    except HTTPError as e:
        return e.code, e.reason
    except URLError as e:
        if isinstance(e.reason, socket.timeout):
            return 408, None
        else:
            return e.reason, None

def POST(uri, post_data:bytes, http_headers= None, time_out= 1, charset= 'UTF-8'):
    """
    http POST method

    - Paramters:
        uri: an uri, it can be an domain or ip, type must is cls<str>.
        data: submit request post data.
        http_headers: set request's headers, default is None.
        time_out: time out for access remote uri , default value is 1 seconds.
        charset: set the http charset, default is UTF-8
    
    - Returns:
        status, body

        example:
            
            200, <html><body>this is an example</body></html>
    """
    try:
        req= None
        if http_headers:
            req= request.Request(uri, data= post_data, headers= http_headers, method= 'POST')
        else:
            req= request.Request(uri, data= post_data, method= 'POST')
        with request.urlopen(req, timeout= time_out) as response:
            return response.getcode(),response.read()
    except HTTPError as e:
        return e.code, e.reason
    except URLError as e:
        if isinstance(e.reason, socket.timeout):
            return 408, None
        else:
            return e.reason, None


from TDhelper.document.excel.meta.modelMeta import modelMeta
class model(metaclass=modelMeta):
    __excelHandle__= None # excel object.
    __sheetHandle__= None # sheet object.
    __rows__= [] # all rows.
    __record_offset__= 1 # data start offset.
    __read_offset__= 0 # now record offset.
    def __init__(self, excelPath= None, record_offset= 1):
        if record_offset<1:
            record_offset=1
        self.__record_offset__= record_offset    
        if excelPath:
            self.Meta.file= excelPath
        if self.Meta.file:
            self.__initExcelHandle__()

    def __initExcelHandle__(self):
        return None

    def __enter__(self):
        return self

    def readLine(self, lineOffset=1):
        '''
            读取一行数据
            - params:
            -   lineOffset: <int>, default 1, 读取的行偏移量.
        '''
        return True
    
    def rows(self):
        '''
            获取表格所有数据.
            - params:
            - return: <arrary>, defautl return [].
        '''
        return []

    def getCount(self):
        '''
            获取总行数
            - return: <int>
        '''
        return 0

    def close(self):
        return None
    
    class Meta:
        '''
            元数据
            - file: <string>, 文件路径.
            - sheet: <string>, sheet名称.
            - extension: <string>, 文件类型(xlsx,csv)
        '''
        file= ''
        sheet= 'sheet1'
        extension= 'xlsx'


#!/usr/bin/env python3.6
# -*- encoding: utf-8 -*-
'''
@File    :   recursion.py
@Time    :   2020/01/16 23:06:51
@Author  :   Tang Jing 
@Version :   1.0.0
@Contact :   yeihizhi@163.com
@License :   (C)Copyright 2020
@Desc    :   None
'''

# here put the import lib
import functools
# code start

def recursion(func):
    def wapper(*args, **kwargs):
        try:
            kwargs['count'] += 1
            if kwargs['count'] == kwargs['upper-limit']:
                return args, kwargs
            return func(*args, **kwargs)
        except Exception:
            pass
    return wapper

def recursionCall(func, upper_limit:int=200, *args, **wargs):
    wargs['count'] = 1
    wargs['upper-limit'] = upper_limit
    wargs['break'] = False
    while True:
        args, wargs = func(*args, **wargs)
        if not wargs['break']:
            wargs['upper-limit'] = wargs['count']+upper_limit
        else:
            break
    return args, wargs


    
if __name__ == "__main__":
    @recursion
    def gggd(*args, **kwargs):
        if kwargs['bbb'] == 3001:
            kwargs['break']= True
            return args, kwargs
        print(kwargs['bbb'])
        kwargs['bbb']+=1
        return gggd(*args, **kwargs)
        
    kwargs={'bbb':1}
    eee=recursionCall(gggd,**kwargs)
    eeee=""


from enum import Enum
from threading import Thread


class actionMode(Enum):
    slow = 0
    middle = 1
    quick = 2


class action:
    def __init__(self):
        self.actions = {}

    def loadActionScript(self, scriptConfig: dict):
        '''
            装载动作脚本
            - params:
            -   scriptConfig: <dict>, 脚本字典. 格式: {"KEY值":脚本对象}
        '''
        if scriptConfig:
            for k, v in scriptConfig.items():
                self.actions[k] = v
        else:
            raise Exception("动作脚本不能为空.")

    def do(self, scriptKey, mode: actionMode):
        '''
            执行动作脚本
            - params:
            -   scriptKey: <str>, 脚本KEY值
            -   mode: <actionMode>, 模式
        '''
        self.actions[scriptKey].do(mode)

    def stop(self, scriptKey):
        '''
            停止执行脚本
            - params:
            -   scriptKey: <str>, 脚本Key值
        '''
        self.actions[scriptKey].stop()

    def stopActionListen(self, scriptKey):
        self.actions[scriptKey].stopListen()



import json

class linked_list:
    '''
    链表
    '''
    def __init__(self, resource: any):
        self.__key= None
        self.__expand_resource= resource
        self.__nodes=[]

    @property
    def key(self):
        return self.__key

    @key.setter
    def key(self, value):
        self.__key= value
    
    @property
    def expand_resource(self):
        return self.__expand_resource
    
    @property
    def nodes(self):
        return self.__nodes
    
    def find(self, find_value:any):
        '''
        查找元素

        - Parameters:
            find_value: 您需要查找的值, 搜索范围(pk_field).
        '''
        # 如果KEY与查找值一致返回对象本身
        if self.__key == find_value:
            return self
        for item in self.__nodes:
            # 进入下级查找
            result= item.find(find_value)
            if result:
                return result
        return None
    
    def toJson(self):
        '''
        cls<node> to json str formatter
        '''
        json_templete= '"{}":&lt;"expand":&lt;{}&gt;,"nodes":['
        ret= json_templete.format(self.__key, self.__getExpand_str())
        if self.__nodes:
            v_count= 1
            for item in self.__nodes:
                str_ret= ''
                str_ret+= item.toJson()
                if v_count < len(self.__nodes):
                    str_ret+= ","
                ret+= "{"+str_ret+"}"
                if v_count < len(self.__nodes):
                    ret+= ","
                v_count+= 1

            ret+= "]}"
        else:
            ret+= "]}"
        return ret
    
    def __getExpand_str(self):
        '''
        generate json's expand  value.
        '''
        ret_str= ''
        v_pos= 0
        for item in self.__expand_resource.__dict__:
            if v_pos > 0:
                ret_str+= ', '
            if isinstance(self.__expand_resource.__dict__[item], str):
                ret_str+= '"'+ str(item) +'" : "' + str(self.__expand_resource.__dict__[item]) + '"'
            else:
                ret_str+= '"'+ str(item) +'" : ' + str(self.__expand_resource.__dict__[item])
            v_pos+= 1
        return ret_str

class Directory:
    '''
    目录结构
    '''
    def __init__(self, pk_field: str, parent_field: str):
        '''
        init

        - Parameters:
            pk_field: 索引字段, cls<string>
            parent_field: 关联索引的字段, cls<string>
        '''
        self.__nodes= []
        self.__pk_field= pk_field
        self.__parent_field= parent_field
    
    def append(self, node: linked_list):
        '''
        增加一个节点
        
        - Parameters:
            node: 节点对象, cls<linked_list> 
        '''
        if node:
            if hasattr(node.expand_resource, self.__pk_field):
                node.key= getattr(node.expand_resource, self.__pk_field)
                if hasattr(node.expand_resource, self.__parent_field):
                    find_value= getattr(node.expand_resource, self.__parent_field)
                    if bool(find_value):
                        for item in self.__nodes:
                            find_ret= item.find(find_value)
                            if find_ret:
                                find_ret.nodes.append(node)
                                return True
                        return False
                    else:
                        self.__nodes.append(node)
                        return True
                else:
                    raise Exception(self.__parent_field + ' can not found.')
            else:
                raise Exception(self.__pk_field + ' can not found.')
        else:
            raise Exception('node is null.')

    def toJson(self):
        '''
        json字符串

        - Returns:
            cls<str>
        '''
        ret= ""
        v_count=1
        for item in self.__nodes:
            ret+= item.toJson() 
            if v_count < len(self.__nodes):
                ret+= ","
            v_count+= 1
        ret= ret.replace('&lt;', '{')
        ret= ret.replace('&gt;', '}')
        ret= ret.replace(',}',"}")
        return "{"+ret+"}"

    def toSerialize(self):
        '''
        序列化

        - Returns:
            cls<dict>
        '''
        return json.loads(self.toJson)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
from io import StringIO

import urllib3


'''
class \r\n
    core.contentExtraction.lineblock\r\n
description\r\n
    行块分析法(内容)\r\n
attribute\r\n
    private __Blocks__\r\n
        type:int\r\n
        default:0\r\n
        description:行块大小\r\n
    private __ChangeRate__\r\n
        type:float\r\n
        default:0.0\r\n
        description:骤变率\r\n
    private __MinLength__\r\n
        type:int\r\n
        default:0\r\n
        description:最小长度\r\n
    private __MapBlock__\r\n
        type:dict()\r\n
        default:{}\r\n
        description:行块列表\r\n
'''
class lineblock:
    __Blocks__=0
    __ChangeRate__=0.0
    __MinLength__=0 
    __MapBlock__=dict()

    def __init__(self,bl_size,cr,mlength):
        '''
        Feature\r\n
            __init__(self,bl_size,cr,mlength)\r\n
        Description\r\n
            初始化\r\n
        Args\r\n
            bl_size\r\n
                type:int\r\n
                description:块偏移\r\n
            cr\r\n
                type:float\r\n
                description:骤变率\r\n
            mlength\r\n
                type:int\r\n
                description:行最小长度\r\n
        '''
        self.__Blocks__=bl_size
        self.__ChangeRate__=cr
        self.__MinLength__=mlength
    
    def createBlocks(self,content):
        '''
        Feature\r\n
            createBlocks(self,content)\r\n
        Description\r\n
            建立块\r\n
        Args\r\n
            content\r\n
                type:string\r\n
                description:内容\r\n
        '''
        ctext=self.filter_tags(content)
        f=StringIO(ctext)
        try:
            #过滤不必要HTML内容生成纯文本
            blockline='' #行块内容 
            blockcount=0   #行块号
            thecount=0     #计数器
            re_line=re.compile(r'[\r\n]',re.I | re.M) #过滤换行
            while True:
                line=f.readline()
                if line=='':
                    break
                line=re_line.sub('',line) #过滤换行
                if len(line)<=self.__MinLength__ and len(blockline)/2-len(line)<=self.__MinLength__:
                    #TODO
                    if thecount<self.__Blocks__:
                        blockline=blockline+line
                        thecount=thecount+1
                    else:
                        if blockline:
                            self.__MapBlock__[blockcount]=blockline
                            blockcount=blockcount+1
                        thecount=0
                        blockline=line
                else:
                    #TODO
                    if blockline:
                        self.__MapBlock__[blockcount]=blockline
                        blockcount=blockcount+1
                    thecount=0
                    blockline=line
                print(line)
                print("-----------------------")
            #处理文本末尾字符
            if thecount!=0:
                self.__MapBlock__[blockcount+1]=blockline
            return True
        except:
            return False
        finally:
            f.flush()
            f.close()

    def getBody(self,ctext):
        '''
        Feature\r\n
            getBody(self,ctext)\r\n
        Description\r\n
            获取内容\r\n
        Args\r\n
            ctext\r\n
                type:string\r\n
                description:需要分析的内容\r\n
        '''
        if self.createBlocks(ctext):
            try:
                currentBlock=len(self.__MapBlock__[0])
                lastBlock=0
                contentBlock={}
                #计算骤变率
                for currentItem in self.__MapBlock__:
                    lastBlock=currentBlock
                    currentBlock=len(self.__MapBlock__[currentItem])
                    between=abs(currentBlock-lastBlock)/max(currentBlock,lastBlock)
                    if currentBlock>self.__MinLength__:
                        contentBlock[currentItem]=between
                    else:
                        if between >= self.__ChangeRate__:
                            contentBlock[currentItem]=between
                #获取正文
                if contentBlock:
                    resultItem=None
                    maxLen=0
                    for Item in contentBlock:
                        if len(self.__MapBlock__[Item])>maxLen:
                            maxLen=len(self.__MapBlock__[Item])
                            resultItem=Item
                    if resultItem:
                        if maxLen>=self.__MinLength__*1.5:
                            return self.__MapBlock__[resultItem]
                        else:
                            return ""
                    else:
                        return ""
                else:
                    return ""
            except:
                pass
            finally:
                pass
        else:
           pass

    def filter_tags(self,htmlstr):
        '''
        Feature\r\n
            filter_tags(self,htmlstr)\r\n
        Description\r\n
            过滤标签\r\n
        Args\r\n
            htmlstr:\r\n
                type:string\r\n
                description:HTML源码\r\n
        '''
        #先过滤CDATA
        re_cdata=re.compile('//<!\[CDATA\[[^>]*//\]\]>',re.I) #匹配CDATA
        re_script=re.compile(r'<script.*?>[\s\S]*?</script*.?>', re.I | re.M)#Script
        re_style=re.compile('<\s*style[^>]*>[^<]*<\s*/\s*style\s*>',re.I)#style
        re_br=re.compile('<br\s*?/?>')#处理换行
        re_comment=re.compile('<!--[^>]*?-->',re.I | re.M)#HTML注释
        re_bian=re.compile(r'<!--[\s\S]+?-->', re.I | re.M)#去除编译条件
        re_h=re.compile('</?\w+[^>]*?>')#HTML标签
        re_stopwords=re.compile('\u3000')#去除无用的'\u3000'字符
        re_tabs=re.compile(r'[\f|\t|\v]',re.I | re.M)#去除制表符等；
        re_tags=re.compile(r'<(\S*?)[^>]*>.*?|<.*? />',re.I | re.M) #过滤剩余的TAGS
        blank_line=re.compile(r'[\r\n]+',re.I | re.M)
        re_space=re.compile(r' ')#去除空格
        s=re_script.sub('',htmlstr)
        s=re_style.sub('',s)#去掉style
        s=re_cdata.sub('',s)#去掉CDATA
        s=re_br.sub('',s)#将br转换空
        s=re_comment.sub('',s)#去掉HTML注释
        s=re_bian.sub('',s)#去除编译条件
        s=re_h.sub('',s) #去掉HTML 标签
        s=re_space.sub('',s)#去掉空格
        s=re_stopwords.sub('',s)#去掉多余的空行
        s=re_tabs.sub('',s)#去掉多余制表符等
        s=blank_line.sub('\n',s)
        s=self.replaceCharEntity(s)#替换实体
        s=re_tags.sub('',s) #过滤剩余TAGS
        return s

    ##替换常用HTML字符实体.
    #使用正常的字符替换HTML中特殊的字符实体.
    #你可以添加新的实体字符到CHAR_ENTITIES中,处理更多HTML字符实体.
    #@param htmlstr HTML字符串.
    def replaceCharEntity(self,htmlstr):
        '''
        Feature\r\n
            replaceCharEntity(self,htmlstr)\r\n
        Description\r\n
            替换HTML字符(特殊字符)\r\n
        Args\r\n
            htmlstr\r\n
                type:string\r\n
                description:要过滤的字符\r\n
        '''
        CHAR_ENTITIES={'nbsp':' ','160':' ',
                    'lt':'<','60':'<',
                    'gt':'>','62':'>',
                    'amp':'&','38':'&',
                    'quot':'"','34':'"',}

        re_charEntity=re.compile(r'&#?(?P<name>\w+);')
        sz=re_charEntity.search(htmlstr)
        while sz:
            entity=sz.group()#entity全称，如&gt;
            key=sz.group('name')#去除&;后entity,如&gt;为gt
            try:
                htmlstr=re_charEntity.sub(CHAR_ENTITIES[key],htmlstr,1)
                sz=re_charEntity.search(htmlstr)
            except KeyError:
                #以空串代替
                htmlstr=re_charEntity.sub('',htmlstr,1)
                sz=re_charEntity.search(htmlstr)
        return htmlstr
    def getPageType(self,args,domain):
        re_href=re.compile('href=\".*?\"',re.I)
        re_link=re.compile('<link.*.?>',re.I)
        args=re_link.sub("",args)
        return "list",""

'''
test fun\r\n
'''
def example:
    from TDhelper.generic.http_helper import m_http
    g=m_http()
    b=lineblock(5,0.2,50)
    content,statucode=g.getcontent("https://3w.huanqiu.com/a/bdb047/9CaKrnKq0iw?agt=8")
    print(content)
    c=b.getBody(content)
    print("正文",c)